/*
  +----------------------------------------------------------------------+
  | PHP Version 5                                                        |
  +----------------------------------------------------------------------+
  | Copyright (c) 1997-2004 The PHP Group                                |
  +----------------------------------------------------------------------+
  | This source file is subject to version 3.0 of the PHP license,       |
  | that is bundled with this package in the file LICENSE, and is        |
  | available through the world-wide-web at the following url:           |
  | http://www.php.net/license/3_0.txt.                                  |
  | If you did not receive a copy of the PHP license and are unable to   |
  | obtain it through the world-wide-web, please send a note to          |
  | license@php.net so we can mail you a copy immediately.               |
  +----------------------------------------------------------------------+
  | Author: Wez Furlong <wez@php.net>                                    |
  | Portions influenced by libevent, written by                          |
  |             Niels Provos <provos@citi.umich.edu>                     |
  +----------------------------------------------------------------------+
*/

/* $Id: event.c,v 1.7 2004/12/16 05:30:09 wez Exp $ */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "php.h"
#include "php_ini.h"
#include "ext/standard/info.h"
#include "php_event.h"
#include "php_event_int.h"

#if HAVE_SIGACTION && HAVE_FCNTL_H
# include <fcntl.h>
#endif

#if HAVE_SOCKETPAIR
# include <sys/socket.h>
#endif

/* True global resources - no need for thread safety here */
static int le_event;

#define event_from_zval(e, z) ZEND_FETCH_RESOURCE(e, struct php_event*, &z, -1, "event", le_event);

/* {{{ event_functions[] */
function_entry event_functions[] = {
	PHP_FE(event_init, NULL)
	PHP_FE(event_have_events, NULL)
	PHP_FE(event_dispatch, NULL)
	PHP_FE(event_new, NULL)
	PHP_FE(event_set, NULL)
	PHP_FE(event_schedule, NULL)
	PHP_FE(event_deschedule, NULL)
	PHP_FE(event_priority_set, NULL)
	PHP_FE(event_pending, NULL)
	PHP_FE(event_free, NULL)
	PHP_FE(event_handle_signal, NULL)
	PHP_FE(event_timeout, NULL)
	{NULL, NULL, NULL}
};
/* }}} */

/* {{{ event_module_entry */
zend_module_entry event_module_entry = {
#if ZEND_MODULE_API_NO >= 20010901
	STANDARD_MODULE_HEADER,
#endif
	"event",
	event_functions,
	PHP_MINIT(event),
	PHP_MSHUTDOWN(event),
	NULL,
	PHP_RSHUTDOWN(event),
	PHP_MINFO(event),
#if ZEND_MODULE_API_NO >= 20010901
	"0.9.1",
#endif
	STANDARD_MODULE_PROPERTIES
};
/* }}} */

#ifdef COMPILE_DL_EVENT
ZEND_GET_MODULE(event)
#endif

static void php_event_init_globals(zend_event_globals *event_globals)
{
	memset(event_globals, 0, sizeof(*event_globals));
}

static void free_event(struct php_event *ev TSRMLS_DC)
{
	if (ev->php_cb) {
		zval_ptr_dtor(&ev->php_cb);
	}
	if (ev->php_cb_arg) {
		zval_ptr_dtor(&ev->php_cb_arg);
	}
	if (ev->php_stream) {
		zend_list_delete(Z_RESVAL_P(ev->php_stream));
		ZVAL_NULL(ev->php_stream);
		zval_ptr_dtor(&ev->php_stream);
	}
	efree(ev);
}

static ZEND_RSRC_DTOR_FUNC(event_dtor) {
	struct php_event *ev = rsrc->ptr;
	event_del(ev TSRMLS_CC);
	free_event(ev TSRMLS_CC);
}

/* {{{ PHP_MINIT_FUNCTION */
PHP_MINIT_FUNCTION(event)
{
	ZEND_INIT_MODULE_GLOBALS(event, php_event_init_globals, NULL);
	le_event = zend_register_list_destructors_ex(event_dtor, NULL, "event", module_number);

	REGISTER_LONG_CONSTANT("EV_TIMEOUT", EV_TIMEOUT, CONST_CS|CONST_PERSISTENT);
	REGISTER_LONG_CONSTANT("EV_READ", EV_READ, CONST_CS|CONST_PERSISTENT);
	REGISTER_LONG_CONSTANT("EV_WRITE", EV_WRITE, CONST_CS|CONST_PERSISTENT);
	REGISTER_LONG_CONSTANT("EV_EXCEP", EV_EXCEP, CONST_CS|CONST_PERSISTENT);
	REGISTER_LONG_CONSTANT("EV_SIGNAL", EV_SIGNAL, CONST_CS|CONST_PERSISTENT);
	REGISTER_LONG_CONSTANT("EV_PERSIST", EV_PERSIST, CONST_CS|CONST_PERSISTENT);

	REGISTER_LONG_CONSTANT("EVLOOP_ONCE", EVLOOP_ONCE, CONST_CS|CONST_PERSISTENT);
	REGISTER_LONG_CONSTANT("EVLOOP_NONBLOCK", EVLOOP_NONBLOCK, CONST_CS|CONST_PERSISTENT);

	return SUCCESS;
}
/* }}} */

/* {{{ PHP_MSHUTDOWN_FUNCTION */
PHP_MSHUTDOWN_FUNCTION(event)
{
	return SUCCESS;
}
/* }}} */

/* {{{ PHP_RSHUTDOWN_FUNCTION */
PHP_RSHUTDOWN_FUNCTION(event)
{
	if (EVENT_G(evsel)) {
		EVENT_G(evsel)->fini(EVENT_G(evbase) TSRMLS_CC);
	}
	return SUCCESS;
}
/* }}} */

static const struct php_eventop *engines[] = {
#ifdef WIN32
	&php_event_win32_engine,
#endif
#if HAVE_WORKING_KQUEUE
	&php_event_kqueue_engine,
#endif
#if HAVE_EPOLL
	&php_event_epoll_engine,
#endif
#if HAVE_SYS_DEVPOLL_H
	&php_event_devpoll_engine,
#endif
#if defined(HAVE_SYS_POLL_H) && defined(HAVE_POLL)
	&php_event_poll_engine,
#endif
	&php_event_select_engine,
	NULL
};

/* {{{ PHP_MINFO_FUNCTION */
PHP_MINFO_FUNCTION(event)
{
	php_info_print_table_start();
	php_info_print_table_header(2, "Event Scheduler support", "enabled");
	php_info_print_table_end();
}
/* }}} */


static int php_event_compare(const struct php_event *a, const struct php_event *b)
{
	if (timercmp(&a->ev_timeout, &b->ev_timeout, <))
		return -1;
	else if (timercmp(&a->ev_timeout, &b->ev_timeout, >))
		return 1;
	return a - b;
}

RB_PROTOTYPE(event_tree, php_event, ev_timeout_node, php_event_compare);
RB_GENERATE(event_tree, php_event, ev_timeout_node, php_event_compare);

#if HAVE_SIGACTION
static void signal_cb(int fd, short mask, void *arg TSRMLS_DC)
{
	int sig[32];
	struct php_event *ev;
	int n, i;

	/* drain any pending signals */
	do {
		n = read(fd, &sig, sizeof(sig));
		for (i = 0; i < n / sizeof(int); i++) {
			TAILQ_FOREACH(ev, &EVENT_G(signalqueue), ev_signal_next) {
				if (ev->ev_fd == sig[i]) {
					event_active(ev, EV_SIGNAL, 1 TSRMLS_CC);
				}
			}
		}
	} while (n == sizeof(sig));
}
#endif

/* {{{ proto string event_init([int priorities [, string engine]])
   Initializes a specified scheduler if engine is specified, else picks the "best" available */
PHP_FUNCTION(event_init)
{
	int i;
	long npriorities = 1;
	char *engine = NULL;
	int engine_len;
	
	if (EVENT_G(evsel)) {
		php_error_docref(NULL TSRMLS_CC, E_WARNING, "You've already initialized using the %s engine",
			EVENT_G(evsel)->name);
		RETURN_FALSE;
	}

	if (FAILURE == zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "|ls", &npriorities, &engine, &engine_len)) {
		RETURN_FALSE;
	}

	gettimeofday(&EVENT_G(event_tv), NULL);
	RB_INIT(&EVENT_G(timetree));
	TAILQ_INIT(&EVENT_G(eventqueue));
#if HAVE_SIGACTION
	TAILQ_INIT(&EVENT_G(signalqueue));
#endif

	if (engine) {
		for (i = 0; engines[i]; i++) {
			if (!strcasecmp(engines[i]->name, engine)) {
				EVENT_G(evbase) = engines[i]->init(TSRMLS_C);
				if (EVENT_G(evbase)) {
					EVENT_G(evsel) = engines[i];
				}
				break;
			}
		}
	} else {
		for (i = 0; engines[i]; i++) {
			EVENT_G(evbase) = engines[i]->init(TSRMLS_C);
			if (EVENT_G(evbase)) {
				EVENT_G(evsel) = engines[i];
				break;
			}
		}
	}

	if (!EVENT_G(evsel)) {
		php_error_docref(NULL TSRMLS_CC, E_WARNING, "Failed to init scheduler");
		RETURN_FALSE;
	}

	/* Allocate our priority queues */
	EVENT_G(nactivequeues) = npriorities;
	EVENT_G(activequeues) = (struct event_list**)ecalloc(npriorities, npriorities * sizeof(struct event_list*));

	for (i = 0; i < npriorities; i++) {
		EVENT_G(activequeues)[i] = emalloc(sizeof(struct event_list));
		TAILQ_INIT(EVENT_G(activequeues)[i]);
	}

#if HAVE_SIGACTION
	sigemptyset(&EVENT_G(evmask));

# if HAVE_SOCKETPAIR
	if (0 == socketpair(AF_UNIX, SOCK_STREAM, 0, EVENT_G(signal_pair)))
# elif HAVE_PIPE
	if (0 == pipe(EVENT_G(signal_pair)))
# endif
	{
# ifdef F_SETFD
		/* close on exec */
		fcntl(EVENT_G(signal_pair)[0], F_SETFD, 1);
		fcntl(EVENT_G(signal_pair)[1], F_SETFD, 1);

		fcntl(EVENT_G(signal_pair)[0], F_SETFL, O_NONBLOCK | fcntl(EVENT_G(signal_pair)[0], F_GETFL, 0));
		fcntl(EVENT_G(signal_pair)[1], F_SETFL, O_NONBLOCK | fcntl(EVENT_G(signal_pair)[1], F_GETFL, 0));
# endif
		/* add one end to the event system */
		event_set(&EVENT_G(sigevent), EVENT_G(signal_pair)[0], EV_PERSIST|EV_READ, signal_cb, &EVENT_G(sigevent) TSRMLS_CC);
		event_add(&EVENT_G(sigevent), NULL TSRMLS_CC);
	}
#endif

	RETURN_STRING(EVENT_G(evsel)->name, 1);
}
/* }}} */

/* {{{ proto bool event_have_events(void)
   Returns true if there are any scheduled events, false otherwise */
PHP_FUNCTION(event_have_events)
{
	RETURN_BOOL(event_haveevents());
}
/* }}} */

static int timeout_next(struct timeval *tv TSRMLS_DC)
{
    struct timeval dflt = TIMEOUT_DEFAULT;

    struct timeval now;
    struct php_event *ev;

    if ((ev = RB_MIN(event_tree, &EVENT_G(timetree))) == NULL) {
        *tv = dflt;
        return (0);
    }

    if (gettimeofday(&now, NULL) == -1)
        return (-1);

    if (timercmp(&ev->ev_timeout, &now, <=)) {
        timerclear(tv);
        return (0);
    }

    timersub(&ev->ev_timeout, &now, tv);

    return (0);
}

static void timeout_correct(struct timeval *off TSRMLS_DC)
{
	struct php_event *ev;
	/*
	 * We can modify the key element of the node without destroying
	 * the key, beause we apply it to all in the right order.
	 */
	RB_FOREACH(ev, event_tree, &EVENT_G(timetree)) {
		timersub(&ev->ev_timeout, off, &ev->ev_timeout);
	}
}

static void timeout_process(TSRMLS_D)
{
    struct timeval now;
    struct php_event *ev, *next;

    gettimeofday(&now, NULL);

    for (ev = RB_MIN(event_tree, &EVENT_G(timetree)); ev; ev = next) {
        if (timercmp(&ev->ev_timeout, &now, >))
            break;
        next = RB_NEXT(event_tree, &EVENT_G(timetree), ev);
        
		event_queue_remove(ev, EVLIST_TIMEOUT TSRMLS_CC);
        /* delete this event from the I/O queues */
		event_del(ev TSRMLS_CC);

        event_active(ev, EV_TIMEOUT, 1 TSRMLS_CC);
    }
}

int event_add(struct php_event *ev, struct timeval *tv TSRMLS_DC)
{
	const struct php_eventop *evsel = EVENT_G(evsel);
	void *evbase = EVENT_G(evbase);

	if (tv != NULL) {
		struct timeval now;

		if (ev->ev_flags & EVLIST_TIMEOUT)
			event_queue_remove(ev, EVLIST_TIMEOUT TSRMLS_CC);

		/* Check if it is active due to a timeout.  Rescheduling
		 * this timeout before the callback can be executed
		 * removes it from the active list. */
		if ((ev->ev_flags & EVLIST_ACTIVE) &&
				(ev->ev_res & EV_TIMEOUT)) {
			/* See if we are just active executing this
			 * event in a loop
			 */
			if (ev->ev_ncalls && ev->ev_pncalls) {
				/* Abort loop */
				*ev->ev_pncalls = 0;
			}

			event_queue_remove(ev, EVLIST_ACTIVE TSRMLS_CC);
		}
		gettimeofday(&now, NULL);
		timeradd(&now, tv, &ev->ev_timeout);

		event_queue_insert(ev, EVLIST_TIMEOUT TSRMLS_CC);
	}

	if ((ev->ev_events & (EV_READ|EV_WRITE)) &&
			!(ev->ev_flags & (EVLIST_INSERTED|EVLIST_ACTIVE))) {
		event_queue_insert(ev, EVLIST_INSERTED TSRMLS_CC);

		return evsel->add(evbase, ev TSRMLS_CC);
	} else if ((ev->ev_events & EV_SIGNAL) &&
			!(ev->ev_flags & EVLIST_SIGNAL)) {
#if HAVE_SIGACTION
		event_queue_insert(ev, EVLIST_SIGNAL TSRMLS_CC);
		sigaddset(&EVENT_G(evmask), ev->ev_fd);
		EVENT_G(signeedsrecalc) = 1;
		return 0;
#else
		php_error_docref(NULL TSRMLS_CC, E_WARNING, "signals are not supported on this platform");
		return -1;
#endif
	}

	return 0;
}

int event_del(struct php_event *ev TSRMLS_DC)
{
	const struct php_eventop *evsel = EVENT_G(evsel);
	void *evbase = EVENT_G(evbase);

	/* See if we are just active executing this event in a loop */
	if (ev->ev_ncalls && ev->ev_pncalls) {
		/* Abort loop */
		*ev->ev_pncalls = 0;
	}

	if (ev->ev_flags & EVLIST_TIMEOUT)
		event_queue_remove(ev, EVLIST_TIMEOUT TSRMLS_CC);

	if (ev->ev_flags & EVLIST_ACTIVE)
		event_queue_remove(ev, EVLIST_ACTIVE TSRMLS_CC);

	if (ev->ev_flags & EVLIST_INSERTED) {
		event_queue_remove(ev, EVLIST_INSERTED TSRMLS_CC);
		return evsel->del(evbase, ev TSRMLS_CC);

	} else if (ev->ev_flags & EVLIST_SIGNAL) {
		event_queue_remove(ev, EVLIST_SIGNAL TSRMLS_CC);
		return evsel->del(evbase, ev TSRMLS_CC);
	}
	return 0;
}

void event_active(struct php_event *ev, int res, short ncalls TSRMLS_DC)
{
	if (!(ev->ev_events & EV_PERSIST)) {
		event_del(ev TSRMLS_CC);
	}

	/* We get different kinds of events, add them together */
	if (ev->ev_flags & EVLIST_ACTIVE) {
		ev->ev_res |= res;
		return;
	}

	ev->ev_res = res;
	ev->ev_ncalls = ncalls;
	ev->ev_pncalls = NULL;
	event_queue_insert(ev, EVLIST_ACTIVE TSRMLS_CC);
}

void event_set(struct php_event *ev, int fd, short events, php_event_callback callback, void *arg TSRMLS_DC)
{
	ev->ev_callback = callback;
	ev->ev_arg = arg;
#ifdef WIN32
	memset(&ev->overlap, 0, sizeof(ev->overlap));
#endif
	ev->ev_fd = fd;
	ev->ev_events = events;
	ev->ev_flags = EVLIST_INIT;
	ev->ev_ncalls = 0;
	ev->ev_pncalls = NULL;

	/* by default, we put new events into the middle priority */
	ev->ev_pri = EVENT_G(nactivequeues)/2;
}

/*
 * Set's the priority of an event - if an event is already scheduled
 * changing the priority is going to fail.
 */
int event_priority_set(struct php_event *ev, int pri TSRMLS_DC)
{
    if (ev->ev_flags & EVLIST_ACTIVE)
        return -1;
    if (pri < 0 || pri >= EVENT_G(nactivequeues))
        return -1;

    ev->ev_pri = pri;

    return 0;
}

/*
 * Checks if a specific event is pending or scheduled.
 */

int event_pending(struct php_event *ev, short event, struct timeval *tv)
{
	int flags = 0;

	if (ev->ev_flags & EVLIST_INSERTED)
		flags |= (ev->ev_events & (EV_READ|EV_WRITE));
	if (ev->ev_flags & EVLIST_ACTIVE)
		flags |= ev->ev_res;
	if (ev->ev_flags & EVLIST_TIMEOUT)
		flags |= EV_TIMEOUT;
	if (ev->ev_flags & EVLIST_SIGNAL)
		flags |= EV_SIGNAL;

	event &= (EV_TIMEOUT|EV_READ|EV_WRITE|EV_SIGNAL);

	/* See if there is a timeout that we should report */
	if (tv != NULL && (flags & event & EV_TIMEOUT))
		*tv = ev->ev_timeout;

	return (flags & event);
}

void event_queue_remove(struct php_event *ev, int queue TSRMLS_DC)
{
	int docount = 1;

	if (!(ev->ev_flags & queue)) {
		php_error_docref(NULL TSRMLS_CC, E_ERROR,
			"event_queue_remove: %p(fd %d) not on queue %x", 
			ev, ev->ev_fd, queue);
	}

	if (ev->ev_flags & EVLIST_INTERNAL) {
		docount = 0;
	}

	if (docount) {
		EVENT_G(event_count)--;
	}

	ev->ev_flags &= ~queue;
	switch (queue) {
		case EVLIST_ACTIVE:
			if (docount) {
				EVENT_G(event_count_active)--;
			}
			TAILQ_REMOVE(EVENT_G(activequeues)[ev->ev_pri],
					ev, ev_active_next);
			break;
		case EVLIST_SIGNAL:
#if HAVE_SIGACTION
			TAILQ_REMOVE(&EVENT_G(signalqueue), ev, ev_signal_next);
#else
			php_error_docref(NULL TSRMLS_CC, E_WARNING, "signals are not supported on this platform");
#endif
			break;
		case EVLIST_TIMEOUT:
			RB_REMOVE(event_tree, &EVENT_G(timetree), ev);
			break;
		case EVLIST_INSERTED:
			TAILQ_REMOVE(&EVENT_G(eventqueue), ev, ev_next);
			break;
		default:
			php_error_docref(NULL TSRMLS_CC, E_ERROR, "event_queue_remove: unknown queue %x", queue);
	}
}

void event_queue_insert(struct php_event *ev, int queue TSRMLS_DC)
{
	int docount = 1;

	if (ev->ev_flags & queue) {
		/* Double insertion is possible for active events */
		if (queue & EVLIST_ACTIVE)
			return;

		php_error_docref(NULL TSRMLS_CC, E_ERROR, "event_queue_insert: %p(fd %d) already on queue %x", 
				ev, ev->ev_fd, queue);
	}

	if (ev->ev_flags & EVLIST_INTERNAL) {
		docount = 0;
	}

	if (docount) {
		EVENT_G(event_count)++;
	}

	ev->ev_flags |= queue;
	switch (queue) {
		case EVLIST_ACTIVE:
			if (docount) {
				EVENT_G(event_count_active)++;
			}
			TAILQ_INSERT_TAIL(EVENT_G(activequeues)[ev->ev_pri],
					ev,ev_active_next);
			break;
		case EVLIST_SIGNAL:
#if HAVE_SIGACTION
			TAILQ_INSERT_TAIL(&EVENT_G(signalqueue), ev, ev_signal_next);
#else
			php_error_docref(NULL TSRMLS_CC, E_WARNING, "signals are not supported on this platform");
#endif
			break;
		case EVLIST_TIMEOUT:
			{
				struct php_event *tmp = RB_INSERT(event_tree, &EVENT_G(timetree), ev);
				assert(tmp == NULL);
			}
			break;
		case EVLIST_INSERTED:
			TAILQ_INSERT_TAIL(&EVENT_G(eventqueue), ev, ev_next);
			break;
		default:
			php_error_docref(NULL TSRMLS_CC, E_ERROR, "event_queue_insert: unknown queue %x", queue);
	}
}


/* Active events are stored in priority queues. Lower priority values are
 * processed before higher priority values and can starve higher priority values. */
static void
event_process_active(TSRMLS_D)
{
	struct php_event *ev;
	struct event_list *activeq = NULL;
	int i;
	short ncalls;

	if (!EVENT_G(event_count_active))
		return;

	for (i = 0; i < EVENT_G(nactivequeues); ++i) {
		if (TAILQ_FIRST(EVENT_G(activequeues)[i]) != NULL) {
			activeq = EVENT_G(activequeues)[i];
			break;
		}
	}

	for (ev = TAILQ_FIRST(activeq); ev; ev = TAILQ_FIRST(activeq)) {
		event_queue_remove(ev, EVLIST_ACTIVE TSRMLS_CC);

		/* Allows deletes to work */
		ncalls = ev->ev_ncalls;
		ev->ev_pncalls = &ncalls;
		while (ncalls) {
			ncalls--;
			ev->ev_ncalls = ncalls;
			(*ev->ev_callback)((int)ev->ev_fd, ev->ev_res, ev->ev_arg TSRMLS_CC);
		}

		/* re-add the event if its a persistent timeout only event */
		if (ev->ev_events == (EV_TIMEOUT|EV_PERSIST)) {
			event_add(ev, &ev->ev_timeout_interval TSRMLS_CC);
		}
	}
}

#if HAVE_SIGACTION
static void signal_handler(int sig)
{
	int n;
	TSRMLS_FETCH();

	/* wake up the scheduling engine */
	n = write(EVENT_G(signal_pair)[1], &sig, sizeof(sig));
}
#endif

/* {{{ proto int event_dispatch([int flags])
   Dispatches events */
PHP_FUNCTION(event_dispatch)
{
	const struct php_eventop *evsel = EVENT_G(evsel);
	void *evbase = EVENT_G(evbase);
	struct timeval tv; 
	int res, done;
	long flags = 0;
	

	if (FAILURE == zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "|l", &flags)) {
		RETURN_LONG(-1);
	}
	
	/* Calculate the initial events that we are waiting for */
	if (evsel->recalc(evbase, 0 TSRMLS_CC) == -1) {
		RETURN_LONG(-1);
	}

#if HAVE_SIGACTION
	if (TAILQ_FIRST(&EVENT_G(signalqueue)) || EVENT_G(signeedsrecalc)) {
		EVENT_G(signeedsrecalc) = 0;

		if (sigprocmask(SIG_BLOCK, &EVENT_G(evmask), NULL) == 0) {
			struct sigaction sa;
			struct php_event *ev;

			memset(&sa, 0, sizeof(sa));
			sa.sa_handler = signal_handler;
			sa.sa_mask = EVENT_G(evmask);
			sa.sa_flags = SA_RESTART;

			TAILQ_FOREACH(ev, &EVENT_G(signalqueue), ev_signal_next) {
				sigaction(ev->ev_fd, &sa, NULL);
			}
		}
	}
#endif

	done = 0;
	while (!done) {
		/* Terminate the loop if we have been asked to */
		if (EVENT_G(got_term)) {
			EVENT_G(got_term) = 0;
			break;
		}

		/* Check if time is running backwards */
		gettimeofday(&tv, NULL);
		if (timercmp(&tv, &EVENT_G(event_tv), <)) {
			struct timeval off;
			timersub(&EVENT_G(event_tv), &tv, &off);
			timeout_correct(&off TSRMLS_CC);
		}
		EVENT_G(event_tv) = tv;

		if (!EVENT_G(event_count_active) && !(flags & EVLOOP_NONBLOCK)) {
			timeout_next(&tv TSRMLS_CC);
		} else {
			timerclear(&tv);
		}

		/* If we have no real events, we just exit */
		if (event_haveevents() == 0) {
			res = 1;
			goto out;
		}

#if HAVE_SIGACTION
		if (TAILQ_FIRST(&EVENT_G(signalqueue))) {
			sigprocmask(SIG_UNBLOCK, &EVENT_G(evmask), NULL);
		}
#endif
		res = evsel->dispatch(evbase, &tv TSRMLS_CC);

		if (res == -1) {
			goto out;
		}

		timeout_process(TSRMLS_C);

		if (EVENT_G(event_count_active)) {
			event_process_active(TSRMLS_C);
			if (!EVENT_G(event_count_active) && (flags & EVLOOP_ONCE))
				done = 1;
		} else if (flags & EVLOOP_NONBLOCK)
			done = 1;

		if (evsel->recalc(evbase, 0 TSRMLS_CC) == -1) {
			res = -1;
			goto out;
		}
	}
	res = 0;
out:
	RETURN_LONG(res);
}
/* }}} */

static void callback_to_php(int fd, short events, void *arg TSRMLS_DC)
{
	struct php_event *ev = (struct php_event*)arg;
	zval *retval = NULL;
	zval **args[3];
	zval *zevents;
	int res;

	args[0] = &ev->php_stream;
	
	MAKE_STD_ZVAL(zevents);
	ZVAL_LONG(zevents, events);
	args[1] = &zevents;

	args[2] = &ev->php_cb_arg;

	res = call_user_function_ex(EG(function_table), NULL, ev->php_cb, &retval, 3, args, 0, NULL TSRMLS_CC);

	if (retval) {
		zval_ptr_dtor(&retval);
	}

	zval_ptr_dtor(&zevents);

	if (res != SUCCESS) {
		php_error_docref(NULL TSRMLS_CC, E_WARNING, "callback failed");
	}	
}

static int event_set_for_php(int isnew, struct php_event *event, zval *zstream, long events, zval *cb, zval *cbarg TSRMLS_DC)
{
	int fd = -1;
	char *callable = NULL;

	if (isnew < 0) {
		fd = -isnew;
		isnew = 1;
	}

	if (event->php_stream) {
		zend_list_delete(Z_RESVAL_P(event->php_stream));
		ZVAL_NULL(event->php_stream);
	}

	if (zstream) {
		php_stream *stream;
		php_stream_from_zval_no_verify(stream, &zstream);
		if (!stream) {
			php_error_docref(NULL TSRMLS_CC, E_WARNING, "Not a valid stream");
			return 0;
		}
		if (FAILURE == php_stream_cast(stream, PHP_STREAM_AS_FD_FOR_SELECT, (void*)&fd, 1) || fd == -1) {
			php_error_docref(NULL TSRMLS_CC, E_WARNING, "Unable to use this stream in the event scheduler");
			return 0;
		}
		if (!event->php_stream) MAKE_STD_ZVAL(event->php_stream);
		ZVAL_RESOURCE(event->php_stream, Z_RESVAL_P(zstream));
		zend_list_addref(Z_RESVAL_P(zstream));

		if (isnew) {
			php_stream_set_option(stream, PHP_STREAM_OPTION_BLOCKING, 0, NULL);
		}
	} else {
		MAKE_STD_ZVAL(event->php_stream);
		if (fd > 0) {
			ZVAL_LONG(event->php_stream, fd);
		}
	}
	
	if (event->php_cb) {
		zval_ptr_dtor(&event->php_cb);
		event->php_cb = NULL;
	}
	if (event->php_cb_arg) {
		zval_ptr_dtor(&event->php_cb_arg);
		event->php_cb_arg = NULL;
	}

#ifdef ZEND_ENGINE_2
	if (!zend_make_callable(cb, &callable TSRMLS_CC)) {
		php_error_docref(NULL TSRMLS_CC, E_WARNING, "function '%s' is not a valid callback", callable);
		efree(callable);
		return 0;
	}
#endif
	
	MAKE_STD_ZVAL(event->php_cb);
	*(event->php_cb) = *cb;
	zval_copy_ctor(event->php_cb);
	
	MAKE_STD_ZVAL(event->php_cb_arg);
	if (cbarg) {
		*(event->php_cb_arg) = *cbarg;
		zval_copy_ctor(event->php_cb_arg);
	}
	
	event_set(event, fd, events, callback_to_php, event TSRMLS_CC);
	
	return 1;
}

/* {{{ proto event event_handle_signal(int signalnum, mixed callback, mixed arg)
   Create a new event to handle signals */
PHP_FUNCTION(event_handle_signal)
{
	zval *zstream = NULL;
	long signo = 0;
	zval *cb = NULL, *cbarg = NULL;
	struct php_event *event;

	if (FAILURE == zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "lz|z!",
			&signo, &cb, &cbarg)) {
		RETURN_FALSE;
	}

	event = ecalloc(1, sizeof(*event));

	if (!event_set_for_php(-signo, event, NULL, EV_SIGNAL|EV_PERSIST, cb, cbarg TSRMLS_CC)) {
		free_event(event TSRMLS_CC);
		RETURN_FALSE;
	}

	/* register the event as a resource, but artificially bump the refcount
	 * so that events created from within functions don't vanish when the
	 * function returns */
	ZEND_REGISTER_RESOURCE(return_value, event, le_event);
	zend_list_addref(Z_RESVAL_P(return_value));
}
/* }}} */


/* {{{ proto event event_new(resource stream, int events, mixed callback, mixed arg)
   Create a new event and set it */
PHP_FUNCTION(event_new)
{
	zval *zstream = NULL;
	long events = 0;
	zval *cb = NULL, *cbarg = NULL;
	struct php_event *event;

	if (FAILURE == zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "r!lz|z!",
			&zstream, &events, &cb, &cbarg)) {
		RETURN_FALSE;
	}

	event = ecalloc(1, sizeof(*event));

	if (!event_set_for_php(1, event, zstream, events, cb, cbarg TSRMLS_CC)) {
		free_event(event TSRMLS_CC);
		RETURN_FALSE;
	}
	
	/* register the event as a resource, but artificially bump the refcount
	 * so that events created from within functions don't vanish when the
	 * function returns */
	ZEND_REGISTER_RESOURCE(return_value, event, le_event);
	zend_list_addref(Z_RESVAL_P(return_value));
}
/* }}} */

/* {{{ proto bool event_set(resource event, resource stream, int events, mixed callback, mixed arg)
   Create a new event and set it */
PHP_FUNCTION(event_set)
{
	zval *zevent = NULL;
	zval *zstream = NULL;
	long events = 0;
	zval *cb = NULL, *cbarg = NULL;
	struct php_event *event;

	if (FAILURE == zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "rr!lz|z!",
			&zevent, &zstream, &events, &cb, &cbarg)) {
		RETURN_FALSE;
	}

	event_from_zval(event, zevent);

	if (!event_set_for_php(0, event, zstream, events, cb, cbarg TSRMLS_CC)) {
		RETURN_FALSE;
	}
	
	RETURN_TRUE;
}
/* }}} */

/* {{{ proto bool event_schedule(resource event [, int timeout_secs [, int timeout usecs]])
   Schedules an event with an optional timeout */
PHP_FUNCTION(event_schedule)
{
	zval *zevent = NULL;
	long ts = -1, tu = -1;
	struct php_event *event;
	struct timeval tv, *tvp = NULL;

	if (FAILURE == zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "r|ll", &zevent, &ts, &tu)) {
		RETURN_FALSE;
	}

	event_from_zval(event, zevent);

	if (ZEND_NUM_ARGS() > 1) {
		tv.tv_sec = ts;
		tv.tv_usec = 0;
		if (ZEND_NUM_ARGS() > 2) {
			tv.tv_usec = tu;
		}
		tvp = &tv;
	}

	RETURN_BOOL(0 == event_add(event, tvp TSRMLS_CC));
}
/* }}} */

/* {{{ proto bool event_deschedule(resource event)
   Deschedule an event; it can then be re-set and re-added */
PHP_FUNCTION(event_deschedule)
{
	zval *zevent;
	struct php_event *event;

	if (FAILURE == zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "r", &zevent)) {
		RETURN_FALSE;
	}

	event_from_zval(event, zevent);
	event->ev_events &= ~EV_PERSIST;

	RETURN_BOOL(0 == event_del(event TSRMLS_CC));
}
/* }}} */

/* {{{ proto void event_free(resource event)
   Destroy an event */
PHP_FUNCTION(event_free)
{
	zval *zevent;
	struct php_event *event;

	if (FAILURE == zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "r", &zevent)) {
		RETURN_FALSE;
	}

	event_from_zval(event, zevent);

	/* double destroy, since we keep an extra ref */
	zend_list_delete(Z_RESVAL_P(zevent));
	zend_list_delete(Z_RESVAL_P(zevent));
}
/* }}} */

/* {{{ proto bool event_priority_set(resource event, int pri)
   Change the priority of an event that is not currently scheduled */
PHP_FUNCTION(event_priority_set)
{
	zval *zevent;
	struct php_event *event;
	long pri;

	if (FAILURE == zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "rl", &zevent, &pri)) {
		RETURN_FALSE;
	}

	event_from_zval(event, zevent);

	RETURN_BOOL(0 == event_priority_set(event, pri TSRMLS_CC));
}
/* }}} */

/* {{{ proto int event_pending(resource event, long mask [, int timeout_secs [, int timeout_usecs]])
   Check if an event is pending or scheduled; returns potential event mask */
PHP_FUNCTION(event_pending)
{
	zval *zevent = NULL;
	long ts = -1, tu = -1, mask;
	struct php_event *event;
	struct timeval tv, *tvp = NULL;

	if (FAILURE == zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "rl|ll", &zevent, &mask, &ts, &tu)) {
		RETURN_FALSE;
	}

	event_from_zval(event, zevent);

	if (ZEND_NUM_ARGS() > 2) {
		tv.tv_sec = ts;
		tv.tv_usec = 0;
		if (ZEND_NUM_ARGS() > 3) {
			tv.tv_usec = tu;
		}
		tvp = &tv;
	}

	RETURN_LONG(event_pending(event, mask, tvp));
}
/* }}} */
   
/* {{{ proto event event_timeout(mixed callback, mixed cbarg, int seconds [, int usecs, [ int flags]])
   Arranges for callback to be called after a certain time delay */
PHP_FUNCTION(event_timeout)
{
	long mask = 0;
	zval *cb = NULL, *cbarg = NULL;
	struct php_event *event;
	long secs, usecs = 0;
	struct timeval tv;

	if (FAILURE == zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "zz!l|ll",
			&cb, &cbarg, &secs, &usecs, &mask)) {
		RETURN_FALSE;
	}

	event = ecalloc(1, sizeof(*event));

	if (!event_set_for_php(1, event, NULL, EV_TIMEOUT | mask, cb, cbarg TSRMLS_CC)) {
		free_event(event TSRMLS_CC);
		RETURN_FALSE;
	}
	
	/* register the event as a resource, but artificially bump the refcount
	 * so that events created from within functions don't vanish when the
	 * function returns */
	ZEND_REGISTER_RESOURCE(return_value, event, le_event);
	zend_list_addref(Z_RESVAL_P(return_value));

	/* schedule it */
	tv.tv_sec = secs;
	tv.tv_usec = usecs;
	event->ev_timeout_interval.tv_sec = secs;
	event->ev_timeout_interval.tv_usec = usecs;

	event_add(event, &tv TSRMLS_CC);
}
/* }}} */

/*
 * Local variables:
 * tab-width: 4
 * c-basic-offset: 4
 * End:
 * vim600: noet sw=4 ts=4 fdm=marker
 * vim<600: noet sw=4 ts=4
 */
